// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

use allegro::*;
use allegro_font::*;
use game_state::{Team, GameState, Object, Vec2D, succubus_animations,
	kiss_animations, devil_hit_animations, stun_animations, bullet_hit_animations, FaceDirection, Collision, CollisionClass};
use engine::id_map::HasId;
use errors::Res;
use DT;
use util::{min, max};
use cgmath::prelude::*;
use ai::{create_heart, create_kiss, create_heart_powerup, create_corpse};
use std::cmp::Ordering;

pub fn create_succubus(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.vel = Some(Vec2D::new(0.0, 0.0));
	obj.old_pos = Some(pos);
	obj.animations = Some(succubus_animations(state)?);
	obj.face_direction = Some(FaceDirection::Right);
	obj.is_player = true;
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
	obj.directional_movement = true;
	obj.collision_class = CollisionClass::Other;
	obj.draw_layer = 1;
	obj.speed = state.options.succubus_speed;
	obj.num_hearts = state.options.succubus_hearts;
	
	Ok(obj)
}

pub fn create_kiss_effect(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.draw_layer = 3;
	obj.time_to_die = state.time + 0.5;
	obj.animations = Some(kiss_animations(state)?);
	Ok(obj)
}

pub fn create_stun_effect(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.draw_layer = 3;
	obj.time_to_die = state.time + 0.5;
	obj.animations = Some(stun_animations(state)?);
	Ok(obj)
}

pub fn create_devil_hit_effect(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.draw_layer = 3;
	obj.time_to_die = state.time + 0.5;
	obj.animations = Some(devil_hit_animations(state)?);
	Ok(obj)
}

pub fn create_bullet_hit_effect(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.draw_layer = 3;
	obj.time_to_die = state.time + 0.5;
	obj.animations = Some(bullet_hit_animations(state)?);
	Ok(obj)
}

pub fn create_damage_effect(parent: usize, pos: Vec2D, damage: i32, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.vel = Some(Vec2D::new(0.0, -55.0));
	obj.time_to_die = state.time + 0.5;
	obj.is_damage_display = true;
	obj.damage = damage;
	
	Ok(obj)
}

pub struct ActorDraw;

complex_behavior!
{
	ActorDraw[obj.animations.is_some() && obj.pos.is_some() && obj.old_pos.is_some()] |self, obj, objs, state|
	{
		let mut y_layer_id = vec![];
		
		for obj in objs.elems()
		{
			if !self.check_object(obj)
			{
				continue;
			}
			let obj = obj.borrow();
			let pos = obj.pos.unwrap();
			y_layer_id.push((pos.y, obj.draw_layer, obj.get_id()));
		}
		
		y_layer_id.sort_by(|&v1, &v2|
		{
			let c1 = v1.1.partial_cmp(&v2.1).unwrap();
			if c1 == Ordering::Equal
			{
				v1.0.partial_cmp(&v2.0).unwrap()
			}
			else
			{
				c1
			}
		});
		
		for (_, _, id) in y_layer_id
		{
			let obj = objs.get(id).unwrap();
			let obj = obj.borrow();
			
			let animations = obj.animations.as_ref().unwrap();
			let pos = obj.pos.as_ref().unwrap();
			let old_pos = obj.old_pos.as_ref().unwrap();
			let moving = obj.moving;
			
			let pos = pos + (pos - old_pos) * state.draw_interp;
			let mut x_offset = 0.0;
			
			let mut animation = None;
			if obj.face_direction.is_some()
			{
				let face_direction = obj.face_direction.unwrap();
				animation = if state.paused
				{
					if animations.dancing.is_some()
					{
						x_offset = 2.0 * (state.core.get_time() * 10.0).sin();
					}
					animations.dancing.as_ref()
				}
				else if obj.kiss_until > state.time 
				{
					match face_direction
					{
						FaceDirection::Left => animations.kiss_left.as_ref(),
						FaceDirection::Down => animations.kiss_down.as_ref(),
						FaceDirection::Up => animations.kiss_up.as_ref(),
						FaceDirection::Right => animations.kiss_right.as_ref(),
					}
				}
				else if moving
				{
					match face_direction
					{
						FaceDirection::Left => animations.walk_left.as_ref(),
						FaceDirection::Down => animations.walk_down.as_ref(),
						FaceDirection::Up => animations.walk_up.as_ref(),
						FaceDirection::Right => animations.walk_right.as_ref(),
					}
				}
				else
				{
					match face_direction
					{
						FaceDirection::Left => animations.left.as_ref(),
						FaceDirection::Down => animations.down.as_ref(),
						FaceDirection::Up => animations.up.as_ref(),
						FaceDirection::Right => animations.right.as_ref(),
					}
				};
			}
			
			let animation = animation.or(animations.default.as_ref());
			
			if let Some(animation) = animation
			{
				animation.draw((pos.x + x_offset as f32).floor(), pos.y.floor(), &state.core);
			}
		}
	}
}

//~ simple_behavior!
//~ {
	//~ CollisionDraw[obj.collision.is_some() && obj.pos.is_some() && obj.old_pos.is_some()] |obj, state|
	//~ {		
		//~ let obj = obj.borrow();
		
		//~ let collision = obj.collision.as_ref().unwrap();
		//~ let pos = obj.pos.as_ref().unwrap();
		//~ let old_pos = obj.old_pos.as_ref().unwrap();
		
		//~ let pos = pos + (pos - old_pos) * state.draw_interp;
		
		//~ let tl = pos + collision.tl;
		//~ let br = pos + collision.br;
		
		//~ state.prim.draw_filled_rectangle(tl.x, tl.y, br.x, br.y, Color::from_rgba_f(0.5, 0.25, 0.25, 0.5));
	//~ }
//~ }

simple_behavior!
{
	PlayerInput[obj.is_player && obj.pos.is_some() && obj.face_direction.is_some()] |obj, state|
	{
		if state.paused
		{
			return Ok(());
		}
		
		let mut obj = obj.borrow_mut();
		
		match state.key_down
		{
			Some(KeyCode::Left) => obj.left = true,
			Some(KeyCode::Right) => obj.right = true,
			Some(KeyCode::Up) => obj.up = true,
			Some(KeyCode::Down) => obj.down = true,
			Some(KeyCode::Space) =>
			{
				if state.time > obj.attack_time && obj.num_hearts > 0
				{
					state.play_sound("data/heal_fire.wav")?;
					
					obj.attack_time = state.time + 0.5;
					obj.time_to_move = state.time + 0.5;
					let heart = create_heart(obj.parent, obj.pos.unwrap(), obj.face_direction.unwrap(), state)?;
					obj.kiss_until = state.time + 0.5;
					state.add_object(heart);
					
					obj.num_hearts -= 1;
				}
			},
			Some(KeyCode::LShift) | Some(KeyCode::RShift) =>
			{
				if state.time > obj.attack_time
				{
					state.play_sound("data/kiss.wav")?;
					
					obj.attack_time = state.time + 0.5;
					obj.time_to_move = state.time + 0.5;
					obj.kiss_until = state.time + 0.5;
					let heart = create_kiss(obj.parent, obj.pos.unwrap(), obj.face_direction.unwrap(), state)?;
					state.add_object(heart);
				}
			},
			_ => (),
		}
		match state.key_up
		{
			Some(KeyCode::Left) => obj.left = false,
			Some(KeyCode::Right) => obj.right = false,
			Some(KeyCode::Up) => obj.up = false,
			Some(KeyCode::Down) => obj.down = false,
			_ => (),
		}
	}
}

fn binary_search<F: Fn(f32) -> bool>(mut dx: f32, acceptor: F) -> f32
{
	let mut dx_min = min(0.0, dx);
	let mut dx_max = max(0.0, dx);
	
	for _ in 0..5
	{
		dx = (dx_min + dx_max) / 2.0;
		
		let collided = acceptor(dx);
		
		let shift_min = collided ^ (dx > 0.0);
		if shift_min
		{
			dx_min = dx;
		}
		else
		{
			dx_max = dx;
		}
	}
	if dx > 0.0
	{
		dx = dx_min;
	}
	else
	{
		dx = dx_max;
	}
	
	dx
}

pub struct ActorPhysics;

complex_behavior!
{
	ActorPhysics[obj.pos.is_some() && obj.vel.is_some()] |self, obj, objs, state|
	{
		if state.paused
		{
			return Ok(());
		}
		for obj in objs.elems()
		{
			if !self.check_object(obj)
			{
				continue
			}
			
			{
				let mut obj = obj.borrow_mut();
				if obj.collision.is_none()
				{
					*obj.pos.as_mut().unwrap() += obj.vel.unwrap() * DT as f32;
					//~ println!("{:?}", obj.vel);
					continue;
				}
			}
			
			let mut collided_collisions = vec![];
			{
				let obj = obj.borrow();
				let collision = obj.collision.unwrap();
				let dpos = obj.vel.unwrap() * DT as f32;
				let collision = collision.add_pos(obj.pos.unwrap() + dpos);
				
				for other_obj in objs.elems()
				{
					let other_obj = other_obj.borrow();
					if !obj.collision_class.collides_with(other_obj.collision_class) || other_obj.pos.is_none() || other_obj.collision.is_none() || other_obj.get_id() == obj.get_id()
					{
						continue;
					}
					
					let other_collision = other_obj.collision.unwrap();
					let other_collision = other_collision.add_pos(other_obj.pos.unwrap());
					
					if other_collision.collides_with(&collision)
					{
						collided_collisions.push(other_collision);
					}
				}
			}
			
			let dpos = if collided_collisions.is_empty()
			{
				let obj = obj.borrow();
				obj.vel.unwrap() * DT as f32
			}
			else
			{
				let obj = obj.borrow();
				let pos = obj.pos.unwrap();
				let dpos = obj.vel.unwrap() * DT as f32;
				let collision = obj.collision.unwrap();
				
				let dx = binary_search(dpos.x,
				|dx|
				{
					let new_pos = pos + Vec2D::new(dx, 0.0);
		
					let collision = collision.add_pos(new_pos);
					
					let mut collided = false;
					for other_collision in &collided_collisions
					{
						if other_collision.collides_with(&collision)
						{
							collided = true;
							break;
						}
					}
					return collided;
				});
				
				let dy = binary_search(dpos.y,
				|dy|
				{
					let new_pos = pos + Vec2D::new(0.0, dy);
		
					let collision = collision.add_pos(new_pos);
					
					let mut collided = false;
					for other_collision in &collided_collisions
					{
						if other_collision.collides_with(&collision)
						{
							collided = true;
							break;
						}
					}
					return collided;
				});
				
				Vec2D::new(dx, dy)
			};
			let mut obj = obj.borrow_mut();
			*obj.pos.as_mut().unwrap() += dpos;	
			
		}
	}
}

simple_behavior!
{
	ActorDirectionalMovement[obj.vel.is_some() && obj.face_direction.is_some() && obj.directional_movement] |obj, state|
	{
		let mut obj = obj.borrow_mut();
		
		let mut vel_vec = Vec2D::new(0.0, 0.0);
		
		if state.time > obj.time_to_move && state.time > obj.stunned_until
		{			
			if obj.left
			{
				vel_vec += Vec2D::new(-obj.speed, 0.0);
			}
			if obj.right
			{
				vel_vec += Vec2D::new(obj.speed, 0.0);
			}
			if obj.up
			{
				vel_vec += Vec2D::new(0.0, -obj.speed);
			}
			if obj.down
			{
				vel_vec += Vec2D::new(0.0, obj.speed);
			}
			
			if vel_vec.magnitude() > obj.speed
			{
				vel_vec *= obj.speed / vel_vec.magnitude();
			}
			
			let moving = {
				let face_direction = obj.face_direction.as_mut().unwrap();
				if vel_vec.x < 0.0
				{
					*face_direction = FaceDirection::Left;
					true
				}
				else if vel_vec.x > 0.0
				{
					*face_direction = FaceDirection::Right;
					true
				}
				else if vel_vec.y < 0.0
				{
					*face_direction = FaceDirection::Up;
					true
				}
				else if vel_vec.y > 0.0
				{
					*face_direction = FaceDirection::Down;
					true
				}
				else
				{
					false
				}
			};
			
			obj.moving = moving;
		}
		else
		{
			obj.moving = false;
		}
		*obj.vel.as_mut().unwrap() = vel_vec;
	}
}

simple_behavior!
{
	OldPos[obj.pos.is_some() && obj.old_pos.is_some()] |obj, _state|
	{		
		let mut obj = obj.borrow_mut();
		
		*obj.old_pos.as_mut().unwrap() = obj.pos.unwrap();
	}
}

simple_behavior!
{
	TimeToDie[obj.time_to_die > 0.0] |obj, state|
	{
		let obj = obj.borrow();
		if obj.time_to_die < state.time
		{
			// LOL...
			
			if obj.is_bullet
			{
				if obj.is_kiss
				{
					let effect = create_kiss_effect(obj.parent, obj.pos.unwrap(), state)?;
					state.add_object(effect);
				}
				else
				{
					if obj.team.is_some()
					{
						let effect = match obj.team.unwrap()
						{
							Team::Soldier => create_bullet_hit_effect(obj.parent, obj.pos.unwrap(), state)?,
							Team::Devil => create_devil_hit_effect(obj.parent, obj.pos.unwrap(), state)?,
						};
						state.add_object(effect);
					}
				}
			}
			
			
			state.remove_object(obj.get_id())
		}
	}
}

pub struct HealthDeath;

complex_behavior!
{
	HealthDeath[obj.health.is_some()] |self, obj, objs, state|
	{
		let mut game_mode_id = None;
		let mut devils_died = 0;
		let mut kissed_died = 0;
		let mut soldiers_died = 0;
		
		for obj in objs.elems()
		{
			{
				let obj = obj.borrow();
				if obj.game_mode.is_some()
				{
					game_mode_id = Some(obj.get_id());
				}
			}
			if !self.check_object(obj)
			{
				continue;
			}
			let obj = obj.borrow();
			
			if obj.health.unwrap() <= 0
			{
				if obj.pos.is_some()
				{
					if obj.kissed
					{
						kissed_died += 1;
						let obj = create_heart_powerup(obj.parent, obj.pos.unwrap(), state)?;
						state.add_object(obj);
					}
					if obj.team.is_some()
					{
						match obj.team.unwrap()
						{
							Team::Soldier => 
							{
								state.play_sound("data/soldier_death.wav")?;
								soldiers_died += 1;
							},
							Team::Devil =>
							{
								state.play_sound("data/devil_death.wav")?;
								devils_died += 1;
							},
						}
						let obj = create_corpse(obj.parent, obj.pos.unwrap(), obj.team.unwrap(), state)?;
						state.add_object(obj);
					}
				}
				state.remove_object(obj.get_id());
			}
		}
		
		if let Some(game_mode_id) = game_mode_id
		{
			let obj = objs.get(game_mode_id).unwrap();
			let mut obj = obj.borrow_mut();
			let game_mode = obj.game_mode.as_mut().unwrap();
			
			game_mode.score += devils_died * 200 - soldiers_died * 100 + kissed_died * 600;
		}
	}
}

simple_behavior!
{
	DamageDraw[obj.is_damage_display && obj.pos.is_some() && obj.old_pos.is_some()] |obj, state|
	{
		let obj = obj.borrow();
		
		let pos = obj.pos.unwrap();
		let old_pos = obj.old_pos.unwrap();
		
		let pos = pos + (pos - old_pos) * state.draw_interp;
		
		let color = if obj.damage > 0
		{
			Color::from_rgb(219, 65, 97)
		}
		else
		{
			Color::from_rgb(81, 162, 0)
		};
		
		state.core.draw_text(&state.ui_font, color, pos.x, pos.y, FontAlign::Centre, &format!("{}", -obj.damage));
	}
}

pub struct HeartPickup;

complex_behavior!
{
	HeartPickup[true] |self, _obj, objs, state|
	{
		for obj in objs.elems()
		{
			{
				let obj = obj.borrow();
				if !obj.is_player || !obj.collision.is_some() || !obj.pos.is_some()
				{
					continue;
				}
			}
			for other_obj in objs.elems()
			{
				{
					let obj = obj.borrow();
					let other_obj = other_obj.borrow();
					if obj.get_id() == other_obj.get_id() || !other_obj.is_heart || !other_obj.collision.is_some() || !other_obj.pos.is_some()
					{
						continue;
					}
				}
				
				let mut obj = obj.borrow_mut();
				let other_obj = other_obj.borrow();
				
				let collision = obj.collision.unwrap();
				let collision = collision.add_pos(obj.pos.unwrap());
				
				let other_collision = other_obj.collision.unwrap();
				let other_collision = other_collision.add_pos(other_obj.pos.unwrap());
				
				if collision.collides_with(&other_collision)
				{
					obj.num_hearts += state.options.heart_pickup;
					state.play_sound("data/pickup.wav")?;
					state.remove_object(other_obj.get_id());
				}
			}
		}
	}
}

pub struct BulletDeath;

complex_behavior!
{
	BulletDeath[obj.is_bullet && obj.pos.is_some()] |self, obj, objs, state|
	{
		for obj in objs.elems()
		{
			if !self.check_object(obj)
			{
				continue
			}
			let obj = obj.borrow();
			
			let mut die = false;
			for other_obj in objs.elems()
			{
				if other_obj.borrow().get_id() == obj.get_id()
				{
					continue;
				}
				let mut other_obj = other_obj.borrow_mut();
				
				if other_obj.health.is_none() || other_obj.collision.is_none()
				{
					continue;
				}
				if obj.team.is_some() && other_obj.team.is_some() && !obj.is_kiss
				{
					if obj.team.unwrap() == other_obj.team.unwrap()
					{
						continue;
					}
				}
				
				let collision = Collision{ tl: Vec2D::new(-obj.effect_size, -obj.effect_size), br: Vec2D::new(32.0 + obj.effect_size, 32.0 + obj.effect_size) };
				let collision = collision.add_pos(obj.pos.unwrap());
				
				let other_collision = other_obj.collision.unwrap();
				let other_collision = other_collision.add_pos(other_obj.pos.unwrap());
				if collision.collides_with(&other_collision)
				{
					if obj.is_kiss
					{
						let effect = create_kiss_effect(obj.parent, obj.pos.unwrap(), state)?;
						state.add_object(effect);
						
						if other_obj.team == Some(Team::Devil)
						{
							state.play_sound("data/stun.wav")?;
							other_obj.stunned_until = state.time + 1.0;
							let effect = create_stun_effect(obj.parent, other_obj.pos.unwrap(), state)?;
							state.add_object(effect);
						}
						else
						{
							*other_obj.health.as_mut().unwrap() -= obj.damage;
							other_obj.kissed = true;
							
							let effect = create_damage_effect(obj.parent, other_obj.pos.unwrap() + Vec2D::new(16.0, 0.0), obj.damage, state)?;
							state.add_object(effect);
						}
					}
					else
					{
						if obj.team.is_some()
						{
							let effect = match obj.team.unwrap()
							{
								Team::Soldier =>
								{
									state.play_sound("data/devil_hurt.wav")?;
									create_bullet_hit_effect(obj.parent, other_obj.pos.unwrap(), state)?
								}
								Team::Devil =>
								{
									state.play_sound("data/soldier_hurt.wav")?;
									create_devil_hit_effect(obj.parent, other_obj.pos.unwrap(), state)?
								},
							};
							state.add_object(effect);
						}
						else
						{
							state.play_sound("data/heal.wav")?;
						}
						
						*other_obj.health.as_mut().unwrap() -= obj.damage;
						
						let effect = create_damage_effect(obj.parent, other_obj.pos.unwrap() + Vec2D::new(16.0, 0.0), obj.damage, state)?;
						state.add_object(effect);
					}
					die = true;
				}
				if die
				{
					break;
				}
			}
			if die
			{
				state.remove_object(obj.get_id())
			}
		}
	}
}
