// Copyright 2017 SiegeLord
// Licensed under GPL, see LICENSE for full terms

use game_state::{Team, GameState, Object, Vec2D, soldier_animations,
	soldier_corpse_animations, devil_animations, devil_corpse_animations, heart_animations,
	bullet_animations, FaceDirection, Collision, Animations, CollisionClass};
use engine::id_map::HasId;
use errors::Res;
use cgmath::prelude::*;

pub fn create_corpse(parent: usize, pos: Vec2D, team: Team, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	match team
	{
		Team::Soldier => obj.animations = Some(soldier_corpse_animations(state)?),
		Team::Devil => obj.animations = Some(devil_corpse_animations(state)?),
	}
	Ok(obj)
}

pub fn create_soldier(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.vel = Some(Vec2D::new(0.0, 0.0));
	obj.old_pos = Some(pos);
	obj.animations = Some(soldier_animations(state)?);
	obj.face_direction = Some(FaceDirection::Right);
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
	obj.team = Some(Team::Soldier);
	obj.directional_movement = true;
	obj.health = Some(state.options.soldier_health);
	obj.is_ai = true;
	obj.collision_class = CollisionClass::Other;
	obj.draw_layer = 1;
	obj.speed = state.options.soldier_speed;
	
	Ok(obj)
}

pub fn create_devil(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.vel = Some(Vec2D::new(0.0, 0.0));
	obj.old_pos = Some(pos);
	obj.animations = Some(devil_animations(state)?);
	obj.face_direction = Some(FaceDirection::Right);
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
	obj.team = Some(Team::Devil);
	obj.directional_movement = true;
	obj.health = Some(state.options.devil_health);
	obj.is_ai = true;
	obj.collision_class = CollisionClass::Other;
	obj.draw_layer = 1;
	obj.speed = state.options.devil_speed;
	
	Ok(obj)
}

pub fn create_bullet(parent: usize, pos: Vec2D, direction: FaceDirection, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	let pos = pos + direction.to_vec() * 16.0;
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.vel = Some(direction.to_vec() * 200.0);
	obj.old_pos = Some(pos);
	obj.face_direction = Some(direction);
	obj.animations = Some(bullet_animations(state)?);
	obj.is_bullet = true;
	obj.collision_class = CollisionClass::Projectile;
	obj.team = Some(Team::Soldier);
	obj.time_to_die = state.time + 0.7;
	obj.damage = 1;
	obj.effect_size = 4.0;
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
	obj.draw_layer = 1;
		
	Ok(obj)
}

pub fn create_heart(parent: usize, pos: Vec2D, direction: FaceDirection, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	let pos = pos + direction.to_vec() * 16.0;
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.vel = Some(direction.to_vec() * 200.0);
	obj.old_pos = Some(pos);
	obj.face_direction = Some(direction);
	obj.animations = Some(heart_animations(state)?);
	obj.is_bullet = true;
	obj.time_to_die = state.time + 0.7;
	obj.collision_class = CollisionClass::Projectile;
	obj.damage = -state.options.heart_heal;
	obj.effect_size = 0.0;
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
		
	Ok(obj)
}

pub fn create_kiss(parent: usize, pos: Vec2D, direction: FaceDirection, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	let pos = pos + direction.to_vec() * 16.0;
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.face_direction = Some(direction);
	obj.animations = Some(Animations::new());
	obj.team = Some(Team::Devil);
	obj.is_bullet = true;
	obj.time_to_die = state.time + 0.1;
	obj.damage = 99;
	obj.effect_size = -5.0;
	obj.is_kiss = true;
	obj.draw_layer = 2;
		
	Ok(obj)
}

pub fn create_heart_powerup(parent: usize, pos: Vec2D, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.animations = Some(heart_animations(state)?);
	obj.is_heart = true;
	obj.collision = Some(Collision{ tl: Vec2D::new(8.0, 32.0 - 16.0), br: Vec2D::new(32.0 - 8.0, 32.0) });
		
	Ok(obj)
}

pub fn create_slash(parent: usize, pos: Vec2D, direction: FaceDirection, state: &mut GameState) -> Res<Object>
{
	let mut obj = Object::new(state.id_mint.new_id());
	
	let pos = pos + direction.to_vec() * 16.0;
	obj.parent = parent;
	obj.pos = Some(pos);
	obj.old_pos = Some(pos);
	obj.face_direction = Some(direction);
	obj.animations = Some(Animations::new());
	obj.is_bullet = true;
	obj.team = Some(Team::Devil);
	obj.time_to_die = state.time + 0.1;
	obj.damage = 2;
	obj.effect_size = 8.0;
		
	Ok(obj)
}


pub struct ActorAI;

complex_behavior!
{
	ActorAI[obj.pos.is_some() && obj.team.is_some() && obj.face_direction.is_some() && obj.is_ai] |self, obj, objs, state|
	{
		if state.paused
		{
			return Ok(());
		}
		
		let mut id_goal_id = vec![];
		let mut game_mode_id = 0;
		
		for obj2 in objs.elems()
		{
			let obj = obj2.borrow();
			if obj.game_mode.is_some()
			{
				game_mode_id = obj.get_id();
			}
			
			if !self.check_object(obj2)
			{
				continue
			}
			
			if obj.stunned_until > state.time
			{
				continue;
			}
			
			let mut goal_id = None;
			let mut min_dist = 0.0;
			for other_obj in objs.elems()
			{
				let other_obj = other_obj.borrow();
				if other_obj.team.is_some() && other_obj.team.unwrap() != obj.team.unwrap() && other_obj.pos.is_some() && other_obj.is_ai
				{
					let dist = (obj.pos.unwrap() - other_obj.pos.unwrap()).magnitude();
					if goal_id.is_none()
					{
						goal_id = Some(other_obj.get_id());
						min_dist = dist;
					}
					else if dist < min_dist
					{
						goal_id = Some(other_obj.get_id());
						min_dist = dist;
					}
				}
			}
			id_goal_id.push((obj.get_id(), goal_id));
		}
		
		for (id, goal_id) in id_goal_id
		{		
			let game_obj = objs.get(game_mode_id).unwrap().borrow();
			let game_mode = game_obj.game_mode.as_ref().unwrap();
			
			let mut obj = objs.get(id).unwrap().borrow_mut();
			
			obj.left = false;
			obj.right = false;
			obj.up = false;
			obj.down = false;
			obj.attacking = false;
			
			if goal_id.is_none()
			{
				continue;
			}
			
			let goal_id = goal_id.unwrap();
			let goal_obj = objs.get(goal_id).unwrap().borrow();
			
			let cur_pos = obj.pos.unwrap();
			let goal_pos = goal_obj.pos.unwrap();
			
			let range = match obj.team.unwrap()
			{
				Team::Soldier => 3,
				Team::Devil => 0,
			};
			
			let dpos = goal_pos - cur_pos;
			if game_mode.clear_shot(cur_pos, goal_pos, range) || (dpos.x.abs() / 32.0 < 0.85 && dpos.y.abs() / 32.0 < 0.85)
			{
				{
					let face_direction = obj.face_direction.as_mut().unwrap();
					
					if dpos.x.abs() > dpos.y.abs()
					{
						if dpos.x < 0.0
						{
							*face_direction = FaceDirection::Left;
						}
						else if dpos.x > 0.0
						{
							*face_direction = FaceDirection::Right;
						}
					}
					else
					{
						if dpos.y < 0.0
						{
							*face_direction = FaceDirection::Up;
						}
						else if dpos.y > 0.0
						{
							*face_direction = FaceDirection::Down;
						}
					}
				}
				
				obj.attacking = true;
				
				if state.time > obj.attack_time
				{
					obj.attack_time = state.time + 1.0;
					obj.time_to_move = state.time + 0.5;
					let obj = match obj.team.unwrap()
					{
						Team::Soldier => create_bullet(obj.parent, cur_pos, obj.face_direction.unwrap(), state)?,
						Team::Devil => create_slash(obj.parent, cur_pos, obj.face_direction.unwrap(), state)?,
					};
					state.add_object(obj);
				}
			}
			else
			{
				let goal_pos = game_mode.get_local_move_goal(obj.get_id(), goal_obj.get_id(), cur_pos, goal_pos);
				
				if goal_pos.x > cur_pos.x + 1.0
				{
					obj.right = true;
				}
				else if goal_pos.x < cur_pos.x - 1.0
				{
					obj.left = true;
				}
				if goal_pos.y > cur_pos.y + 1.0
				{
					obj.down = true;
				}
				else if goal_pos.y < cur_pos.y - 1.0
				{
					obj.up = true;
				}
			}
		}
	}
}
